@@ -10,7 +10,9 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # pylint: disable=R0901
+import json
 from collections import UserList, OrderedDict
+from contextlib import contextmanager
 from functools import wraps
 from pprint import pprint
 from time import sleep
@@ -21,6 +23,39 @@ from version_utils import rpm
 from adcm_client.util.search import search_one, search
 from adcm_client.wrappers.api import ADCMApiWrapper
 
+# If we are running the client from tests with Allure we expected that code
+# to trace steps in Allure UI.
+# But in case of running client outside of testing Allure is useless in virtualenv.
+# So that code should be flexible enought to work with Allure or without.
+ALLURE = True
+try:
+    import allure
+except ImportError:
+    ALLURE = False
+
+
+# That is trick which is almost the same that in _allure.py::StepContext
+# We have a function that can be used as contextmanager and decorator
+# in same time.
+@contextmanager
+def dummy_context(*args, **kwargs):
+    yield
+
+
+def allure_step(text):
+    if ALLURE:
+        return allure.step(text)
+    return dummy_context(text)
+
+
+def allure_attach_json(body, name):
+    if ALLURE:
+        allure.attach(
+            json.dumps(body, indent=2),
+            name=name,
+            attachment_type=allure.attachment_type.TEXT,
+        )
+
 
 def pp(*args, **kwargs):
     pprint("--------------------------------------------------")
@@ -81,7 +116,7 @@ class TooOldServerVersion(Exception):
                             f' upgrade version ADCM.')
         else:
             self.message = message
-        super(TooOldServerVersion, self).__init__(self.message)
+        super().__init__(self.message)
 
 
 def min_server_version(version):
@@ -97,6 +132,28 @@ def min_server_version(version):
     return decorate
 
 
+def legacy_server_implementaion(oldfunc, turnover_version):
+    """Replace function with oldfunc if version of server < turnover_version
+
+    This decorator is usefull in case you have two way to do almost the same for
+    new server and for old server. For example you have new, more efficient calls.
+
+    There is a restriction of using this decorator. You should not patch legacy and
+    non legacy functions with setattr. Because we call original class functions from
+    inner wrapper.
+    But any you are not so dirty, don't you?
+    """
+    def decorator(func):
+        @wraps(func)
+        def wrapper(self, *args, **kwargs):
+            # adcm_version >= turnover_versions
+            if rpm.compare_versions(self.adcm_version, turnover_version) < 0:
+                return oldfunc(self, *args, **kwargs)
+            return func(self, *args, **kwargs)
+        return wrapper
+    return decorator
+
+
 class Paging:
     def __init__(self, paged_object, limit=50, **args):
         self._paged_object = paged_object
@@ -123,7 +180,7 @@ class Paging:
             self._current_list = result
             self._current_iterator = iter(self._current_list)
         except PagingEnds:
-            raise StopIteration
+            raise StopIteration from None
 
     def __next_element(self):
         return next(self._current_iterator)
@@ -200,14 +257,16 @@ class EndPoint:
         except coreapi.exceptions.ErrorMessage as e:
             # pylint: disable=W0212
             if "code" in e.error._data and e.error._data["code"] == "TOO_LONG":
-                raise ResponseTooLong
+                raise ResponseTooLong from e
             raise e
 
         if isinstance(result, OrderedDict):
             # It's paging mode
-            if result['results'] == []:
+            if not result['results']:
                 raise PagingEnds
             return result['results']
+        elif isinstance(result, list) and 'offset' in paging and paging['offset']:
+            raise PagingEnds
         return result
 
     def read(self, object_id):
@@ -221,7 +280,11 @@ class EndPoint:
         # FIXME: paging
         if self.idname in args:
             return self.read(args[self.idname])
-        data = search_one(self.list(**args), **args)
+        data = None
+        for obj in Paging(self.list, **args):
+            data = search_one([obj], **args)
+            if data is not None:
+                break
         if data is None:
             raise ObjectNotFound
 
