@@ -9,50 +9,29 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-# pylint: disable=R0901, R0904, W0401, C0302
+# pylint: disable=too-many-lines, too-many-public-methods, too-many-ancestors
 
 import logging
-from contextlib import contextmanager
+import warnings
+from collections import abc
+from json import dumps
 
 from coreapi.exceptions import ErrorMessage
 from version_utils import rpm
 
 from adcm_client.base import (
     ActionHasIssues, ADCMApiError, BaseAPIListObject, BaseAPIObject, ObjectNotFound,
-    TooManyArguments, strip_none_keys, min_server_version
+    TooManyArguments, WaitTimeout, strip_none_keys, min_server_version, allure_step,
+    allure_attach_json, legacy_server_implementaion
 )
 from adcm_client.util import stream
 from adcm_client.wrappers.api import ADCMApiWrapper
 
-# If we are running the client from tests with Allure we expected that code
-# to trace steps in Allure UI.
-# But in case of running client outside of testing Allure is useless in virtualenv.
-# So that code should be flexible enought to work with Allure or without.
-ALLURE = True
-try:
-    import allure
-except ImportError:
-    ALLURE = False
-
 # Init logger
-logger = logging.getLogger()
+logger = logging.getLogger(__name__)
 logger.setLevel(logging.DEBUG)
 
 
-# That is trick which is allmost the same that in _allure.py::StepContext
-# We have a function that can be used as contextmanager and decorator
-# in same time.
-@contextmanager
-def dummy_context(text):
-    yield text
-
-
-def allure_step(text):
-    if ALLURE:
-        return allure.step(text)
-    return dummy_context(text)
-
-
 class NoCredentionsProvided(Exception):
     """There is no user/password provided. It was not passsed as init parameters
     during ADCMClient initialization and was not passed as a parameter to auth()
@@ -71,7 +50,6 @@ class Bundle(BaseAPIObject):
     IDNAME = "bundle_id"
     PATH = ["stack", "bundle"]
     FILTERS = ["name", "version"]
-
     id = None
     bundle_id = None
     name = None
@@ -79,6 +57,9 @@ class Bundle(BaseAPIObject):
     version = None
     edition = None
 
+    def __repr__(self):
+        return f"<Bundle {self.name} {self.version} {self.edition} at {id(self)}>"
+
     def provider_prototype(self) -> "ProviderPrototype":
         return self._child_obj(ProviderPrototype)
 
@@ -86,21 +67,21 @@ class Bundle(BaseAPIObject):
         try:
             prototype = self.provider_prototype()
         except ObjectNotFound:
-            raise IncorrectPrototypeType
+            raise IncorrectPrototypeType from None
         return prototype.provider_create(name, description)
 
     def provider_list(self, paging=None, **args) -> "ProviderList":
         try:
             prototype = self.provider_prototype()
         except ObjectNotFound:
-            raise IncorrectPrototypeType
+            raise IncorrectPrototypeType from None
         return prototype.provider_list(paging=paging, **args)
 
     def provider(self, **args) -> "Provider":
         try:
             prototype = self.provider_prototype()
         except ObjectNotFound:
-            raise IncorrectPrototypeType
+            raise IncorrectPrototypeType from None
         return prototype.provider(**args)
 
     def service_prototype(self, **args) -> "ServicePrototype":
@@ -113,21 +94,21 @@ class Bundle(BaseAPIObject):
         try:
             prototype = self.cluster_prototype()
         except ObjectNotFound:
-            raise IncorrectPrototypeType
+            raise IncorrectPrototypeType from None
         return prototype.cluster_create(name, description)
 
     def cluster_list(self, paging=None, **args) -> "ClusterList":
         try:
             prototype = self.cluster_prototype()
         except ObjectNotFound:
-            raise IncorrectPrototypeType
+            raise IncorrectPrototypeType from None
         return prototype.cluster_list(paging=paging, **args)
 
     def cluster(self, **args) -> "Cluster":
         try:
             prototype = self.cluster_prototype()
         except ObjectNotFound:
-            raise IncorrectPrototypeType
+            raise IncorrectPrototypeType from None
         return prototype.cluster(**args)
 
     def license(self):
@@ -205,11 +186,13 @@ class ServicePrototype(Prototype):
     imports = None
     monitoring = None
 
+    @min_server_version('2020.09.25.13')
     def service_list(self, paging=None, **args) -> "ServiceList":
         return self._child_obj(ServiceList, paging=paging, **args)
 
+    @min_server_version('2020.09.25.13')
     def service(self, **args) -> "Service":
-        return self._child_obj(Service)
+        return self._child_obj(Service, **args)
 
 
 class ServicePrototypeList(BaseAPIListObject):
@@ -285,6 +268,8 @@ class _BaseObject(BaseAPIObject):
         return self._subobject(ActionList, paging=paging, **args)
 
     def action_run(self, **args) -> "Task":
+        warnings.warn('Deprecated. The method accepts no arguments for the "action.run()" method.',
+                      DeprecationWarning, stacklevel=2)
         action = self.action(**args)
         return action.run()
 
@@ -296,29 +281,37 @@ class _BaseObject(BaseAPIObject):
 
     @allure_step("Save config")
     def config_set(self, data):
+        # this check is incomplete, cases of presence of keys "config" and "attr" in config
+        # are not considered
+        allure_attach_json(data, name="Complete config")
         if "config" in data and "attr" in data:
-            # We are in a new mode with full_info == True
             if data["attr"] is None:
-                data["attr"] = []
+                data["attr"] = {}
             history_entry = self._subcall(
-                'config', 'history', 'create',
-                config=data["config"],
-                attr=data["attr"],
-            )
-            return history_entry
+                'config', 'history', 'create', config=data['config'], attr=data['attr'])
+            return {key: value for key, value in history_entry.items() if key in ['config', 'attr']}
         history_entry = self._subcall('config', 'history', 'create', config=data)
         return history_entry['config']
 
     @allure_step("Save config")
     def config_set_diff(self, data):
-        config = self.config()
-        for gk, gv in data.items():
-            if isinstance(gv, dict):
-                for k, v in gv.items():
-                    config[gk][k] = v
-            else:
-                config[gk] = gv
-        self.config_set(config)
+
+        def update(d, u):
+            # if the old and new values are dictionaries, we try to update, otherwise we replace
+            for key, value in u.items():
+                if isinstance(value, abc.Mapping) and key in d and isinstance(d[key], abc.Mapping):
+                    d[key] = update(d[key], value)
+                    continue
+                d[key] = value
+            return d
+
+        # this check is incomplete, cases of presence of keys "config" and "attr" in config
+        # are not considered
+        allure_attach_json(data, name="Changed fields")
+        is_full = "config" in data and "attr" in data
+        config = self.config(full=is_full)
+        allure_attach_json(config, name="Original config")
+        return self.config_set(update(config, data))
 
     def config_prototype(self):
         return self.prototype().config
@@ -336,6 +329,9 @@ class Provider(_BaseObject):
     description = None
     bundle_id = None
 
+    def __repr__(self):
+        return f"<Provider {self.name} at {id(self)}>"
+
     def bundle(self) -> "Bundle":
         return self._parent_obj(Bundle)
 
@@ -382,6 +378,9 @@ class Cluster(_BaseObject):
     serviceprototype = None
     status = None
 
+    def __repr__(self):
+        return f"<Cluster {self.name} form bundle - {self.bundle_id} at {id(self)}>"
+
     def prototype(self) -> "ClusterPrototype":
         return self._parent_obj(ClusterPrototype)
 
@@ -419,18 +418,39 @@ class Cluster(_BaseObject):
         with allure_step("Remove host {} from cluster {}".format(host.fqdn, self.name)):
             self._subcall("host", "delete", host_id=host.id)
 
-    def service(self, **args) -> "Service":
+    def _service_old(self, **args):
         return self._subobject(Service, **args)
 
-    def service_list(self, paging=None, **args) -> "ServiceList":
+    # !!! If you change the version, do not forget to change it in the __new__ method
+    # of the Service class as well as in the comments to them
+    @legacy_server_implementaion(_service_old, '2020.09.25.13')
+    def service(self, **args) -> "Service":
+        return self._child_obj(Service, **args)
+
+    def _service_list_old(self, paging=None, **args):
         return self._subobject(ServiceList, paging=paging, **args)
 
-    def service_add(self, **args) -> "Service":
+    # !!! If you change the version, do not forget to change it in the __new__ method
+    # of the Service class as well as in the comments to them
+    @legacy_server_implementaion(_service_list_old, '2020.09.25.13')
+    def service_list(self, paging=None, **args) -> "ServiceList":
+        return self._child_obj(ServiceList, paging=paging, **args)
+
+    def _service_add_old(self, **args):
         proto = self.bundle().service_prototype(**args)
         with allure_step("Add service {} to cluster {}".format(proto.name, self.name)):
             data = self._subcall("service", "create", prototype_id=proto.id)
             return self._subobject(Service, service_id=data['id'])
 
+    # !!! If you change the version, do not forget to change it in the __new__ method
+    # of the Service class as well as in the comments to them
+    @legacy_server_implementaion(_service_add_old, '2020.09.25.13')
+    def service_add(self, **args) -> "Service":
+        proto = self.bundle().service_prototype(**args)
+        with allure_step("Add service {} to cluster {}".format(proto.name, self.name)):
+            data = self._subcall("service", "create", prototype_id=proto.id, cluster_id=self.id)
+            return Service(self._api, id=data['id'])
+
     @min_server_version('2020.05.13.00')
     def service_delete(self, service: "Service"):
         with allure_step("Remove service {} from cluster {}".format(service.name, self.name)):
@@ -442,6 +462,7 @@ class Cluster(_BaseObject):
     @allure_step("Save hostcomponents map")
     def hostcomponent_set(self, *hostcomponents):
         hc = []
+        readable_hc = []
         for i in hostcomponents:
             h, c = i
             hc.append({
@@ -449,13 +470,19 @@ class Cluster(_BaseObject):
                 'service_id': c.service_id,
                 'component_id': c.id
             })
+            readable_hc.append({
+                'host_fqdn': h.fqdn,
+                'component_name': c.display_name
+            })
+        allure_attach_json(readable_hc, name="Readable hc map")
+        allure_attach_json(hc, name="Complete hc map")
         return self._subcall("hostcomponent", "create", hc=hc)
 
     def status_url(self):
         return self._subcall("status", "list")
 
     def imports(self):
-        raise NotImplementedError
+        return self._subcall("import", "list")
 
     def upgrade(self, **args) -> "Upgrade":
         return self._subobject(Upgrade, **args)
@@ -511,8 +538,8 @@ class UpgradeList(BaseAPIListObject):
 ##################################################
 class Service(_BaseObject):
     IDNAME = "service_id"
-    PATH = None
-    SUBPATH = ["service"]
+    PATH = ['service']
+    SUBPATH = ['service']
 
     id = None
     service_id = None
@@ -525,6 +552,22 @@ class Service(_BaseObject):
     button = None
     monitoring = None
 
+    def __new__(cls, *args, **kwargs):
+        """
+        Set PATH=None, if adcm version < `2020.09.25.13`. See ADCM-1439.
+        This method is associated with the action of the `legacy_server_implementaion()` decorator.
+        """
+        wrapper = args[0]
+        instance = super().__new__(cls)
+        # !!! If you change the version, do not forget to change it in the service(), service_list()
+        # and service_add() methods of the Cluster class as well as in the comments to them
+        if rpm.compare_versions(wrapper.adcm_version, '2020.09.25.13') < 0:
+            instance.PATH = None
+        return instance
+
+    def __repr__(self):
+        return f"<Service {self.name} form cluster - {self.cluster_id} at {id(self)}>"
+
     def bind(self, target):
         if isinstance(target, Cluster):
             self._subcall("bind", "create", export_cluster_id=target.cluster_id)
@@ -537,49 +580,127 @@ class Service(_BaseObject):
     def prototype(self) -> "ServicePrototype":
         return ServicePrototype(self._api, id=self.prototype_id)
 
+    def cluster(self) -> Cluster:
+        return Cluster(self._api, id=self.cluster_id)
+
     def imports(self):
-        raise NotImplementedError
+        return self._subcall("import", "list")
 
     def bind_list(self, paging=None):
         return self._subcall("bind", "list")
 
-    def component(self, **args) -> "Component":
+    def _component_old(self, **args) -> "Component":
         return self._subobject(Component, **args)
 
-    def component_list(self, paging=None, **args) -> "ComponentList":
+    # Set a real version when components feature will be merged into develop
+    # https://github.com/arenadata/adcm/pull/778
+    # !!! If you change the version, do not forget to change it in the __new__ method
+    # of the Component class as well as in the comments to them
+    @legacy_server_implementaion(_component_old, '2021.12.01.01')
+    def component(self, **args) -> "Component":
+        return self._child_obj(Component, **args)
+
+    def _component_list_old(self, paging=None, **args) -> "ComponentList":
         return self._subobject(ComponentList, paging=paging, **args)
 
+    # Set a real version when components feature will be merged into develop
+    # https://github.com/arenadata/adcm/pull/778
+    # !!! If you change the version, do not forget to change it in the __new__ method
+    # of the Component class as well as in the comments to them
+    @legacy_server_implementaion(_component_list_old, '2021.12.01.01')
+    def component_list(self, paging=None, **args) -> "ComponentList":
+        return self._child_obj(ComponentList, paging=paging, **args)
+
 
 class ServiceList(BaseAPIListObject):
-    SUBPATH = ["service"]
+    PATH = ['service']
+    SUBPATH = ['service']
     _ENTRY_CLASS = Service
 
+    def __new__(cls, *args, **kwargs):
+        """
+        Set PATH=None, if adcm version < `2020.09.25.13`. See ADCM-1439.
+        This method is associated with the action of the `legacy_server_implementaion()` decorator.
+        """
+        wrapper = args[0]
+        instance = super().__new__(cls)
+        # !!! If you change the version, do not forget to change it in the service(), service_list()
+        # and service_add() methods of the Cluster class as well as in the comments to them
+        if rpm.compare_versions(wrapper.adcm_version, '2020.09.25.13') < 0:
+            instance.PATH = None
+        return instance
+
 
 ##################################################
 #           C O M P O N E N T S
 ##################################################
-class Component(BaseAPIObject):
+class Component(_BaseObject):
     IDNAME = "component_id"
+    PATH = ["component"]
     SUBPATH = ["component"]
 
     id = None
     component_id = None
+    cluster_id = None
+    _service_id = None
     name = None
     display_name = None
     description = None
     constraint = None
     params = None
     prototype_id = None
+    requires = None
+    bound_to = None
+    monitoring = None
+    status = None
+    state = None
+
+    def __new__(cls, *args, **kwargs):
+        """
+        Set PATH=None, if adcm version < `2021.12.01.01`. See ADCM-1439.
+        This method is associated with the action of the `legacy_server_implementaion()` decorator.
+        """
+        wrapper = args[0]
+        instance = super().__new__(cls)
+        # !!! If you change the version, do not forget to change it in the component()
+        # and component_list() methods of the Service class as well as in the comments to them
+        if rpm.compare_versions(wrapper.adcm_version, '2021.12.01.01') < 0:
+            instance.PATH = None
+        return instance
 
     @property
     def service_id(self):
-        return self._endpoint.path_args["service_id"]
+        # this code is for backward compatibility
+        if self._service_id is not None:
+            return self._service_id
+        try:
+            return self._endpoint.path_args["service_id"]
+        except KeyError:
+            return self._data['service_id']
+
+    @service_id.setter
+    def service_id(self, value):
+        self._service_id = value
 
 
 class ComponentList(BaseAPIListObject):
+    PATH = ["component"]
     SUBPATH = ["component"]
     _ENTRY_CLASS = Component
 
+    def __new__(cls, *args, **kwargs):
+        """
+        Set PATH=None, if adcm version < `2021.12.01.01`. See ADCM-1439.
+        This method is associated with the action of the `legacy_server_implementaion()` decorator.
+        """
+        wrapper = args[0]
+        instance = super().__new__(cls)
+        # !!! If you change the version, do not forget to change it in the component()
+        # and component_list() methods of the Service class as well as in the comments to them
+        if rpm.compare_versions(wrapper.adcm_version, '2021.12.01.01') < 0:
+            instance.PATH = None
+        return instance
+
 
 ##################################################
 #              H O S T
@@ -598,6 +719,9 @@ class Host(_BaseObject):
     bundle_id = None
     status = None
 
+    def __repr__(self):
+        return f"<Host {self.fqdn} form provider - {self.provider_id} at {id(self)}>"
+
     def provider(self) -> "Provider":
         return self._parent_obj(Provider)
 
@@ -650,6 +774,9 @@ class Action(BaseAPIObject):
     subs = None
     config = None
 
+    def __repr__(self):
+        return f"<Action {self.name} at {id(self)}>"
+
     def _get_config(self):
         config = dict()
         for item in self.config['config']:
@@ -673,6 +800,9 @@ class Action(BaseAPIObject):
     def run(self, **args) -> "Task":
         with allure_step("Run action {}".format(self.name)):
 
+            if 'hc' in args:
+                allure_attach_json(args.get('hc'), name="Hostcomponent map")
+
             if 'config' in args and 'config_diff' in args:
                 raise TypeError("only one argument is expected 'config' or 'config_diff'")
 
@@ -688,15 +818,18 @@ class Action(BaseAPIObject):
                         key, subkey = item['name'], item['subname']
                         if subkey and subkey in config_diff.get(key, {}):
                             args['config'][key][subkey] = config_diff[key][subkey]
-                        elif not subkey and config_diff.get(key):
+                        elif not subkey and key in config_diff:
                             args['config'][key] = config_diff[key]
-
+            # check backward compatibility for `verbose` option
+            if 'verbose' not in args and rpm.compare_versions(
+                    self.adcm_version, '2021.02.04.13') >= 0:
+                args['verbose'] = False
             try:
                 data = self._subcall("run", "create", **args)
             except ErrorMessage as error:
                 if (getattr(error.error, 'title', '') == '409 Conflict'
                         and 'has issues' in getattr(error.error, '_data', {}).get('desc', '')):
-                    raise ActionHasIssues
+                    raise ActionHasIssues from error
                 raise error
             return Task(self._api, task_id=data["id"])
 
@@ -713,12 +846,20 @@ class TaskFailed(Exception):
     pass
 
 
+TASK_PARENT = {
+    'cluster': Cluster,
+    'service': Service,
+    'host': Host,
+    'provider': Provider,
+    'component': Component,
+}
+
+
 class Task(BaseAPIObject):
     IDNAME = "task_id"
     PATH = ["task"]
     FILTERS = ['action_id', 'pid', 'status', 'start_date', 'finish_date']
     _END_STATUSES = ["failed", "success"]
-    action = None
     action_id = None
     config = None
     hostcomponentmap = None
@@ -729,6 +870,18 @@ class Task(BaseAPIObject):
     selector = None
     status = None
     url = None
+    object_id = None
+    object_type = None
+
+    @min_server_version('2020.08.27.00')
+    def action(self) -> "Action":
+        # for component object method will work after version `2021.12.01.01`
+        kwargs = {f'{self.object_type}_id': self.object_id}
+        return TASK_PARENT[self.object_type](
+            self._api, **kwargs).action(action_id=self.action_id)
+
+    def __repr__(self):
+        return f"<Task {self.task_id} at {id(self)}>"
 
     def job(self, **args) -> "Job":
         return Job(self._api, path_args=dict(task_id=self.id), **args)
@@ -738,11 +891,16 @@ class Task(BaseAPIObject):
 
     @allure_step("Wait for task end")
     def wait(self, timeout=None, log_failed=True):
-        status = self.wait_for_attr("status",
-                                    self._END_STATUSES,
-                                    timeout=timeout)
-        if log_failed and status == "failed":
-            self._log_jobs('failed')
+        try:
+            status = self.wait_for_attr("status",
+                                        self._END_STATUSES,
+                                        timeout=timeout)
+            if log_failed and status == "failed":
+                self._log_jobs(status=status)
+        except WaitTimeout as e:
+            if log_failed:
+                self._log_jobs()
+            raise WaitTimeout from e
         return status
 
     @allure_step("Wait for task to success.")
@@ -754,12 +912,20 @@ class Task(BaseAPIObject):
 
         return status
 
-    def _log_jobs(self, status):
-        for job in self.job_list(status=status):
+    def _log_jobs(self, **filters):
+        for job in self.job_list(**filters):
+            log_func = logger.error if job.status == "failed" else logger.info
+            log_func("Action: %s", self.action().name)
             for file in job.log_files:
                 response = self._api.client.get(file["url"])
-                if 'content' in response:
-                    logger.error(response["content"])
+                content_format = response.get("format", "txt")
+                if "type" in response:
+                    log_func("Type: %s", response['type'])
+                if "content" in response:
+                    if content_format == "json":
+                        log_func(dumps(response["content"], indent=2))
+                    else:
+                        log_func(response["content"])
 
 
 class TaskList(BaseAPIListObject):
@@ -801,11 +967,20 @@ class Job(BaseAPIObject):
     url = None
     log_files = None
     task_id = None
+    display_name = None
+    start_date = None
+    finish_date = None
+
+    def __repr__(self):
+        return f"<Job {self.job_id} at {id(self)}>"
 
     # FIXME: remove method __init__, deal with argument path_args
     def __init__(self, api: ADCMApiWrapper, path=None, path_args=None, **args):
         super().__init__(api, path, **args)
 
+    def task(self) -> "Task":
+        return self._parent_obj(Task)
+
     def wait(self, timeout=None):
         return self.wait_for_attr("status",
                                   self._END_STATUSES,
@@ -825,7 +1000,7 @@ class JobList(BaseAPIListObject):
 ##################################################
 #              A D C M
 ##################################################
-class ADCM(BaseAPIObject):
+class ADCM(_BaseObject):
     IDNAME = "adcm_id"
     PATH = ["adcm"]
     id = None
@@ -835,48 +1010,15 @@ class ADCM(BaseAPIObject):
     prototype_version = None
     bundle_id = None
 
-    # TODO: Remove that function when it become first class object
-    def config(self, full=False):
-        history_entry = self._subcall("config", "current", "list")
-        if full:
-            return history_entry
-        return history_entry['config']
-
-    @allure_step("Save config")
-    def config_set(self, data):
-        if "config" in data and "attr" in data:
-            # We are in a new mode with full_info == True
-            if data["attr"] is None:
-                data["attr"] = []
-            history_entry = self._subcall(
-                'config', 'history', 'create',
-                config=data["config"],
-                attr=data["attr"],
-            )
-            return history_entry
-        history_entry = self._subcall('config', 'history', 'create', config=data)
-        return history_entry['config']
-
-    @allure_step("Save config")
-    def config_set_diff(self, data):
-        config = self.config()
-        for gk, gv in data.items():
-            if isinstance(gv, dict):
-                for k, v in gv.items():
-                    config[gk][k] = v
-            else:
-                config[gk] = gv
-        self.config_set(config)
-
-    def config_prototype(self):
-        return self.prototype().config
+    def prototype(self) -> "Prototype":
+        return Prototype(self._api, id=self.prototype_id)
 
 
 ##################################################
 #              C L I E N T
 ##################################################
 class ADCMClient:
-    _MIN_VERSION = "2020.04.29.15"
+    _MIN_VERSION = "2019.02.20.00"
 
     def __init__(self, api=None, url=None, user=None, password=None):
         if api is not None:
@@ -888,9 +1030,15 @@ class ADCMClient:
             self.url = url
             self._api = ADCMApiWrapper(self.url)
             self.auth(user, password)
+
         if self.api_token() is not None:
             self.guess_adcm_url()
+
         self.adcm_version = self._api.adcm_version
+        self._check_min_version()
+
+    def __repr__(self):
+        return f"<ADCM API Client for {self.url} at {id(self)}>"
 
     def auth(self, user=None, password=None):
         if user is None or password is None:
@@ -901,8 +1049,10 @@ class ADCMClient:
         self._check_min_version()
 
     def _check_min_version(self):
-        if rpm.compare_versions(self._MIN_VERSION, self._api.adcm_version) > 0:
-            raise ADCMApiError("That client supports ADCM versions >= {}".format(self._MIN_VERSION))
+        if rpm.compare_versions(self._MIN_VERSION, self._api.adcm_version) > -1:
+            raise ADCMApiError(
+                "The client supports ADCM versions newer than '{}'".format(self._MIN_VERSION)
+            )
 
     def api_token(self):
         return self._api.api_token
