@@ -11,7 +11,6 @@
 # limitations under the License.
 # pylint: disable=W0611, W0621, W0404, W0212, C1801
 import coreapi
-import logging
 import requests
 
 try:
@@ -24,9 +23,6 @@ except ImportError:
     IS_ALLURE = False
 
 
-logging.getLogger("urllib3").setLevel(logging.ERROR)
-
-
 class APINode():
     pass
 
@@ -130,13 +126,32 @@ class ADCMApiWrapper():
         except (KeyError, AttributeError):
             self.adcm_version = "0"
 
-    def action(self, *args, **kvargs):
+    def action(self, *args, **kwargs):
         """
         Do operation over api. For information see coreapi documentation.
 
         Example:
         api.action(['cluster', 'create'], name='testcluster')
         """
-        data = self.client.action(self.schema, *args, **kvargs)
+
+        # After parsing the schema, fields for a query, a form, or a path may appear that have
+        # the same names, which may cause the field value to go to the wrong place.
+        # Path fields take precedence over query and form fields, so if there are path fields
+        # and query or form fields with the same name, we delete the query and form fields
+        # with the given name.
+        path = args[0]
+        link = self.schema[path[0]]
+        for item in path[1:]:
+            link = link[item]
+
+        fields = []
+        path_fields = [field.name for field in link.fields if field.location == 'path']
+        for field in link.fields:
+            if not (field.location in ('query', 'form') and field.name in path_fields):
+                fields.append(field)
+        fields = tuple(fields)
+        overrides = {'fields': fields}
+
+        data = self.client.action(self.schema, *args, overrides=overrides, **kwargs)
         self._check_for_error(data)
         return data
